@@ -64,6 +64,9 @@ class DFSDLLWrapper(object):
     @staticmethod
     def dfsReadItemTimeStep(HeaderPointer, FilePointer, Time, Data):
         return hllDll['dfsReadItemTimeStep'](HeaderPointer, FilePointer, byref(Time), Data)
+
+
+
     @staticmethod
     def dfsStaticRead(HeaderPointer, fioError):
         hllDll.dfsStaticRead.argtypes=[c_longlong, POINTER(c_int)]
@@ -326,6 +329,9 @@ class DFSBase(object):
         self.delete_value=-1e-035
         self.projection_WKT = 'NON-UTM'
         self._needs_saving=False
+        self.use_cache=False
+        self._cache = {}
+
     
     @classmethod
     def open_file(cls, file_name):
@@ -349,6 +355,17 @@ class DFSBase(object):
         cls.__open_for_reading()
         return cls
 
+    @classmethod
+    def from_template_file(cls, file_name, TemplateFileName):
+        """
+        Creates a new file copying all info from the Templatefile except the data and the timesteps.
+        """
+        TemplateFile = DFSBase.open_file(TemplateFileName)
+        cls = TemplateFile.new_file(file_name, TemplateFile.number_of_items)
+        cls._copy_item_info(TemplateFile)
+        cls.__copy_from_template(TemplateFile)
+        return cls
+
 
     def __open_for_reading(self):
         """
@@ -429,14 +446,6 @@ class DFSBase(object):
         else:
             self.compression_array = np.zeros((self.number_of_rows, self.number_of_columns,), dtype=int)
 
-
-                #Prepare the array to read data
-        if self.space_axis is SpaceAxisType.EqD2:
-            self.dfsdata=(c_float*self.number_of_columns *self.number_of_rows)()
-        elif self.space_axis is SpaceAxisType.EqD3:
-            self.dfsdata=(c_float*self.number_of_columns *self.number_of_rows*self.number_of_layers)()
-
-
     def new_file(self, FileName, NumberOfItems):
         """
         Creates a new dfs-file. The NumberOfItems has to be given and cannot be changed later
@@ -492,7 +501,7 @@ class DFSBase(object):
                 zkey[i] = 0;
         return xkey, ykey, zkey, en
 
-    def copy_from_template(self, dfs):
+    def __copy_from_template(self, dfs):
         """
         Copies all static info from another dfs-file. Does not copy the data
         """
@@ -507,6 +516,16 @@ class DFSBase(object):
             xkey, ykey, zkey, en = dfs.__read_compression__()
             DFSDLLWrapper().dfsSetEncodeKey(self._header_pointer, xkey, ykey, zkey, en)
 
+        self.number_of_columns = dfs.number_of_columns;
+        self.number_of_rows = dfs.number_of_rows;
+        self.number_of_layers = dfs.number_of_layers
+        self.dx = dfs.dx
+        self.dy = dfs.dy
+        self.x_origin = dfs.x_origin
+        self.y_origin = dfs.y_origin
+        self.orientation = dfs.orientation
+
+
     def get_time_of_timestep(self, TimeStep):
         """
         Gets the time of a particular timestep
@@ -573,8 +592,12 @@ class DFSBase(object):
 
             ltime = c_double(0)
 
-                        #Prepare the array to read data
-            if self.space_axis is not SpaceAxisType.EqD2 and self.space_axis is not SpaceAxisType.EqD3:
+            #Prepare the array to read data
+            if self.space_axis is SpaceAxisType.EqD2:
+                self.dfsdata=(c_float*self.number_of_columns *self.number_of_rows)()
+            elif self.space_axis is SpaceAxisType.EqD3:
+                self.dfsdata=(c_float*self.number_of_columns *self.number_of_rows*self.number_of_layers)()
+            else:
                 self.dfsdata = (c_float*self.items[Item_index-1].number_of_elements)()
 
             self.error = DFSDLLWrapper().dfsReadItemTimeStep(self._header_pointer, self._file_pointer, ltime, self.dfsdata)
@@ -809,10 +832,12 @@ class DFS0(DFSBase):
     def __init__(self, file_name):
         super(DFS0, self).__init__(file_name)
         self.__data_read=False
+        self.use_cache=True
 
     @classmethod
     def from_file(cls, file_name):
         """
+        Deprecated. Use open_file(file_name) instead
         Opens a dfs0 file for reading
         """
         ext=Path(file_name).suffix.lower()
@@ -833,21 +858,22 @@ class DFS0(DFSBase):
 
         return cls
     
-    @classmethod
-    def from_template_file(cls, FileName, TemplateFile):
-        """
-        Creates a new .dfs0-file with the same items as in the templatefile but with out data.
-        """
-        cls = super().new_file(FileName, TemplateFile.NumberOfItems)
-        cls._copy_item_info(TemplateFile)
-        cls.copy_from_template(TemplateFile)
-        return cls
 
     def get_data(self, TimeStep, Item):
         """
         Gets the data for one Time step and one item. Time step counts from 0. Item counts from 1.
         """
-        return super(DFS0, self).read_item_time_step(TimeStep, Item)[0]
+        if Item in self._cache:
+            currentts = self._cache[Item]
+            if TimeStep in currentts:
+                return self._cache[Item][TimeStep] #We have already read this item
+        else:
+            currentts = {}
+            self._cache[Item]=currentts
+        data = super(DFS0, self).read_item_time_step(TimeStep, Item)[0]
+        if self.use_cache:
+            currentts[TimeStep]=data
+        return data
 
     def set_data(self, TimeStep, Item, Data):
         """
@@ -863,13 +889,11 @@ class DFS0(DFSBase):
         Gets all data for a particular item. Reads the entire file at the first call
         """
         if not self.__data_read:
-            for I in self.items:
-                I.data = []
             for t in range(0, self.number_of_time_steps):
                 for inumber in range(1, self.number_of_items +1):
-                    self.items[inumber-1].data.append(self.get_data(t, inumber))
+                    self.get_data(t,inumber)
             self.__data_read = True
-        return self.items[Item].data
+        return list(self._cache[Item].values())
 
 
     def get_timeseries(self, Item):
@@ -901,71 +925,51 @@ class DFS2(DFSBase):
         cls.time_axis = TimeAxisType.CalendarEquidistant
         cls.space_axis = SpaceAxisType.EqD2
         return cls
-    
-    @classmethod
-    def from_template_file(cls, file_name, TemplateFileName):
-        TemplateFile = DFS2.from_file(TemplateFileName)
-        cls = DFS2.new_file(file_name, TemplateFile.number_of_items)
-        cls._copy_item_info(TemplateFile)
-        cls.copy_from_template(TemplateFile)
-        return cls
+           
 
-    def copy_from_template(self, TemplateFile):
-        super(DFS2, self).copy_from_template(TemplateFile)
-        self.number_of_columns = TemplateFile.number_of_columns;
-        self.number_of_rows = TemplateFile.number_of_rows;
-        self.dx = TemplateFile.dx
-        self.dy = TemplateFile.dy
-        self.x_origin = TemplateFile.x_origin
-        self.y_origin = TemplateFile.y_origin
-        self.orientation = TemplateFile.orientation
-        
     def get_data(self, TimeStep, Item):
-        return ma.array(self.read_item_time_step(TimeStep, Item), mask=self.compression_array,fill_value= self.delete_value)
-        return np.array(super(DFS2,self).read_item_time_step(TimeStep, Item))
+        """
+        Gets the data for a TimeStep and Item as masked numpy array.
+        Uses cache if use_cache is set to true
+        """
+        if Item in self._cache:
+            currentts = self._cache[Item]
+            if TimeStep in currentts:
+                return self._cache[Item][TimeStep] #We have already read this item
+        else:
+            currentts = {}
+            self._cache[Item]=currentts
+        data = ma.array(self.read_item_time_step(TimeStep, Item), mask=self.compression_array,fill_value= self.delete_value)
+        if self.use_cache:
+            currentts[TimeStep]=data
+                
+        return data
+
+    def get_values(self, Item, row, column):
+        """
+        Gets all data for a particular item. Reads through the entire file at the first call
+        """
+        toreturn =[]
+        for t in range(0, self.number_of_time_steps):
+            toreturn.append(self.get_data(t, Item)[row, column])
+        return toreturn
+
+
+    def get_timeseries(self, Item, row, column):
+        """
+        Gets all data for a particular item and cell as a Pandas series
+        """
+        return pd.Series(self.get_values(Item, row, column), index=self.time_steps)
+
 
     def to_asc(self, file_name, time_step, item):
         dtpm = np.flipud(self.get_data(time_step, item))
         np.savetxt(file_name, dtpm, fmt = '%10.5E', header= self.get_asc_header(), comments='')
         
-       
     def set_data(self, TimeStep, Item, Data):
         super(DFS2, self).write_item_timestep(TimeStep, Item, Data.ctypes)
 
 
-def get_points_within_polygon(dfs, polygon):
-    """
-    Returns a list of points that are within the polygon. Uses the contains method from matplotlib
-    """
-    from matplotlib import path
-    from matplotlib.transforms import Bbox
-
-    toreturn =[]
-    bbox = polygon.get_extents()
-    col1 = dfs.get_column_index(bbox.intervalx[0])
-    #We go to the boundaries even though the bounding box is outside of the model domain
-    if col1==-1:
-        col1=0
-    col2 = dfs.get_column_index(bbox.intervalx[1])
-    if col2==-2:
-        col2=dfs.number_of_columns
-    row1 = dfs.get_row_index(bbox.intervaly[0])
-    if row1==-1:
-        row1=0
-    row2 = dfs.get_row_index(bbox.intervaly[1])
-    if row2 ==-2:
-        row =dfs.number_of_rows
-    
-    for row in range(row1, row2):
-        for col in range(col1,col2):
-            if dfs.compression_array[row, col]==1:
-                point = DfsPoint(col, row, dfs.get_x_center(col), dfs.get_y_center(row), dfs)
-                if polygon.contains_point((point.x, point.y)):
-                    toreturn.append(point)
-    return toreturn
-
-
-
 class DFS3(DFSBase):
     """
     A class for reading and writing .dfs3-files
@@ -982,44 +986,56 @@ class DFS3(DFSBase):
 
     @classmethod
     def new_file(cls, file_name, NumberOfItems):
+        """
+        Creates a new .dfs3-file
+        """
         cls = DFS3(file_name)
         DFSDLLWrapper().dfsHeaderCreate(FileType.EqtimeFixedspaceAllitems, NumberOfItems, StatType.NoStat, cls._header_pointer)
         super(DFS3, cls).new_file(file_name, NumberOfItems)
         cls.time_axis = TimeAxisType.CalendarEquidistant
         cls.space_axis = SpaceAxisType.EqD3
         return cls
-    
-    @classmethod
-    def from_template_file(cls, FileName, TemplateFileName):
-        TemplateFile = DFS3.from_file(TemplateFileName)
-        cls = DFS3.new_file(FileName, TemplateFile.number_of_items)
-        cls._copy_item_info(TemplateFile)
-        cls.copy_from_template(TemplateFile)
-        return cls
-
-    def copy_from_template(self, TemplateFile):
-        super(DFS3, self).copy_from_template(TemplateFile)
-        self.number_of_columns = TemplateFile.number_of_columns
-        self.number_of_rows = TemplateFile.number_of_rows
-        self.number_of_layers = TemplateFile.number_of_layers
-        self.dx = TemplateFile.dx
-        self.dy = TemplateFile.dy
-        self.x_origin = TemplateFile.x_origin
-        self.y_origin = TemplateFile.y_origin
-        self.orientation = TemplateFile.orientation
-        
+            
     def get_data(self, TimeStep, Item):
         """
         returns a list of numpy arrays with values for each layer. Item counts from 1, Timestep from 0
         """
-        data =super(DFS3, self).read_item_time_step(TimeStep, Item)
+
+        if Item in self._cache:
+            currentts = self._cache[Item]
+            if TimeStep in currentts:
+                return currentts[TimeStep] #We have already read this item
+        else:
+            currentts = {}
+            self._cache[Item]=currentts
+        data=None
+        data = super(DFS3, self).read_item_time_step(TimeStep, Item)
         toreturn =[]
         
         for i in range(self.number_of_layers):
             toreturn.append(ma.array(data[:][:][i], mask=self.compression_array,fill_value= self.delete_value))
-            
-#            toreturn.append(np.array(data[:][:][i]))
+        if self.use_cache:
+            currentts[TimeStep]=toreturn
+                
         return toreturn
+
+    def get_values(self, Item, row, column, Layer):
+        """
+        Gets all data for a particular item. Reads through the entire file at the first call
+        """
+        toreturn =[]
+        for t in range(0, self.number_of_time_steps):
+            toreturn.append(self.get_data(t, Item)[Layer][row, column])
+        return toreturn
+
+
+    def get_timeseries(self, Item, row, column, Layer):
+        """
+        Gets all data for a particular item and cell as a Pandas series
+        """
+        return pd.Series(self.get_values(Item, row, column, Layer), index=self.time_steps)
+
+
         
     def set_data(self, TimeStep, Item, Data):
         """
@@ -1114,6 +1130,37 @@ class DfsPoint(object):
         self.dfsfile =dfsfile
 
 
+def get_points_within_polygon(dfs, polygon):
+    """
+    Returns a list of points that are within the polygon. Uses the contains method from matplotlib
+    """
+    from matplotlib import path
+    from matplotlib.transforms import Bbox
+
+    toreturn =[]
+    bbox = polygon.get_extents()
+    col1 = dfs.get_column_index(bbox.intervalx[0])
+    #We go to the boundaries even though the bounding box is outside of the model domain
+    if col1==-1:
+        col1=0
+    col2 = dfs.get_column_index(bbox.intervalx[1])
+    if col2==-2:
+        col2=dfs.number_of_columns
+    row1 = dfs.get_row_index(bbox.intervaly[0])
+    if row1==-1:
+        row1=0
+    row2 = dfs.get_row_index(bbox.intervaly[1])
+    if row2 ==-2:
+        row =dfs.number_of_rows
+    
+    for row in range(row1, row2):
+        for col in range(col1,col2):
+            if dfs.compression_array[row, col]==1:
+                point = DfsPoint(col, row, dfs.get_x_center(col), dfs.get_y_center(row), dfs)
+                if polygon.contains_point((point.x, point.y)):
+                    toreturn.append(point)
+    return toreturn
+
 #Enumeration section
 class SpaceAxisType(Enum):
     undefined=0
