@@ -255,12 +255,12 @@ class River_setup(object):
                     dg = next(dgp for dgp in r.grid_points if np.float32(dgp.chainage) == dataitem.chainages[ci])
 
                     if dataitem.name == "Water level":
-                        dg.Hitemindex = dataitem.ItemIndex
+                        dg.Hitemindex = dataitem.ItemIndex+1
                         dg.Hindex = ci + dataitem.offset;
                         r.grid_points_h.append(dg)
                         dg.point_type+='H'
                     elif dataitem.name == "Discharge":
-                        dg.itemindex = dataitem.ItemIndex
+                        dg.itemindex = dataitem.ItemIndex+1
                         dg.index = ci + dataitem.offset;
                         r.grid_points_q.append(dg)
                         dg.point_type+='Q'
@@ -406,7 +406,7 @@ class River_setup(object):
         """
         if ( 'Q' in grid_point.point_type):
             #We need to add 1 because we have the item index not the item number
-            return self.res.get_timeseries(grid_point.itemindex+1, grid_point.index)
+            return self.res.get_timeseries(grid_point.itemindex, grid_point.index)
 
     def get_values_Q(self, grid_point):
         """
@@ -423,7 +423,7 @@ class River_setup(object):
         """
         if ( 'H' in grid_point.point_type):
             #We need to add 1 because we have the item index not the item number
-            return self.res.get_timeseries(grid_point.Hitemindex+1, grid_point.Hindex)
+            return self.res.get_timeseries(grid_point.Hitemindex, grid_point.Hindex)
 
 
     def streamstage_ts(self, reach):
